@@ -22,6 +22,7 @@
 #include "theme-thumbnail.h"
 #include "gnome-theme-apply.h"
 #include "theme-installer.h"
+#include "theme-save.h"
 #include "theme-util.h"
 
 #include <glib/gi18n.h>
@@ -30,7 +31,57 @@
 
 #define CUSTOM_THEME_NAME "__custom__"
 
-static void theme_thumbnail_done_cb (GdkPixbuf *pixbuf, AppearanceData *data);
+static void
+theme_thumbnail_done_cb (GdkPixbuf *pixbuf, gchar *theme_name, AppearanceData *data)
+{
+  /* find item in model and update thumbnail */
+  if (pixbuf) {
+    GtkTreeIter iter;
+    GtkTreeModel *model = GTK_TREE_MODEL (data->theme_store);
+
+    if (theme_find_in_model (model, theme_name, &iter))
+      gtk_list_store_set (data->theme_store, &iter, COL_THUMBNAIL, pixbuf, -1);
+
+    g_object_unref (pixbuf);
+  }
+}
+
+static void
+theme_thumbnail_generate (GnomeThemeMetaInfo *info, AppearanceData *data)
+{
+  generate_meta_theme_thumbnail_async (info,
+      (ThemeThumbnailFunc) theme_thumbnail_done_cb, data, NULL);
+}
+
+static void
+theme_changed_on_disk_cb (GnomeThemeType       type,
+			  gpointer             theme,
+			  GnomeThemeChangeType change_type,
+			  GnomeThemeElement    element,
+			  AppearanceData       *data)
+{
+  if (type == GNOME_THEME_TYPE_METATHEME) {
+    GnomeThemeMetaInfo *meta = theme;
+
+    if (change_type == GNOME_THEME_CHANGE_CREATED) {
+      gtk_list_store_insert_with_values (data->theme_store, NULL, 0,
+          COL_LABEL, meta->readable_name,
+          COL_NAME, meta->name,
+          COL_THUMBNAIL, data->theme_icon,
+          -1);
+      theme_thumbnail_generate (meta, data);
+
+    } else if (change_type == GNOME_THEME_CHANGE_DELETED) {
+      GtkTreeIter iter;
+
+      if (theme_find_in_model (GTK_TREE_MODEL (data->theme_store), meta->name, &iter))
+        gtk_list_store_remove (data->theme_store, &iter);
+
+    } else if (change_type == GNOME_THEME_CHANGE_CHANGED) {
+      theme_thumbnail_generate (meta, data);
+    }
+  }
+}
 
 static gchar *
 get_default_string_from_key (GConfClient *client, const char *key)
@@ -56,83 +107,20 @@ is_locked_down (GConfClient *client)
   return gconf_client_get_bool (client, LOCKDOWN_KEY, NULL);
 }
 
-static gboolean
-find_in_model (GtkTreeModel *model, const gchar *value, gint column, GtkTreeIter *hit)
-{
-  GtkTreeIter iter;
-  gboolean valid;
-  gchar *test;
-
-  if (!value)
-    return FALSE;
-
-  for (valid = gtk_tree_model_get_iter_first (model, &iter); valid;
-       valid = gtk_tree_model_iter_next (model, &iter))
-  {
-    gtk_tree_model_get (model, &iter, column, &test, -1);
-
-    if (test) {
-      gint cmp = strcmp (test, value);
-      g_free (test);
-
-      if (!cmp) {
-      	if (hit)
-          *hit = iter;
-        return TRUE;
-      }
-    }
-  }
-
-  return FALSE;
-}
-
 static void
 theme_load_from_gconf (GConfClient *client, GnomeThemeMetaInfo *theme)
 {
-  gchar *s;
+  g_free (theme->gtk_theme_name);
+  theme->gtk_theme_name = gconf_client_get_string (client, GTK_THEME_KEY, NULL);
 
-  s = gconf_client_get_string (client, GTK_THEME_KEY, NULL);
-  if (s != NULL) {
-    g_free (theme->gtk_theme_name);
-    theme->gtk_theme_name = s;
-  }
+  g_free (theme->gtk_color_scheme);
+  theme->gtk_color_scheme = gconf_client_get_string (client, COLOR_SCHEME_KEY, NULL);
 
-  s = gconf_client_get_string (client, COLOR_SCHEME_KEY, NULL);
-  if (s != NULL) {
-    g_free (theme->gtk_color_scheme);
-    theme->gtk_color_scheme = s;
-  }
+  g_free (theme->metacity_theme_name);
+  theme->metacity_theme_name = gconf_client_get_string (client, METACITY_THEME_KEY, NULL);
 
-  s = gconf_client_get_string (client, METACITY_THEME_KEY, NULL);
-  if (s != NULL) {
-    g_free (theme->metacity_theme_name);
-    theme->metacity_theme_name = s;
-  }
-
-  s = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
-  if (s != NULL) {
-    g_free (theme->icon_theme_name);
-    theme->icon_theme_name = s;
-  }
-}
-
-static gboolean
-theme_thumbnail_generate (AppearanceData *data)
-{
-  generate_theme_thumbnail_async (data->theme_queue->data,
-      (ThemeThumbnailFunc) theme_thumbnail_done_cb, data, NULL);
-  return FALSE;
-}
-
-static void
-theme_queue_for_thumbnail (GnomeThemeMetaInfo *theme, AppearanceData *data)
-{
-  gboolean idle = (data->theme_queue == NULL);
-
-  data->theme_queue = g_slist_append (data->theme_queue, theme);
-
-  if (idle)
-    theme_thumbnail_generate (data);
+  g_free (theme->icon_theme_name);
+  theme->icon_theme_name = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
 }
 
 static gchar *
@@ -192,19 +180,23 @@ theme_select_name (GtkIconView *icon_view, const gchar *theme)
   GtkTreeIter iter;
   GtkTreeModel *model = gtk_icon_view_get_model (icon_view);
 
-  if (find_in_model (model, theme, COL_NAME, &iter))
+  if (theme_find_in_model (model, theme, &iter))
     theme_select_iter (icon_view, &iter);
 }
 
 static gboolean
 theme_is_equal (const GnomeThemeMetaInfo *a, const GnomeThemeMetaInfo *b)
 {
+  gboolean a_set, b_set;
+
   if (!(a->gtk_theme_name && b->gtk_theme_name) ||
       strcmp (a->gtk_theme_name, b->gtk_theme_name))
     return FALSE;
 
-  if (!(a->gtk_color_scheme && b->gtk_color_scheme) ||
-      strcmp (a->gtk_color_scheme, b->gtk_color_scheme))
+  a_set = a->gtk_color_scheme && strcmp (a->gtk_color_scheme, "");
+  b_set = b->gtk_color_scheme && strcmp (b->gtk_color_scheme, "");
+  if ((a_set != b_set) ||
+      (a_set && strcmp (a->gtk_color_scheme, b->gtk_color_scheme)))
     return FALSE;
 
   if (!(a->icon_theme_name && b->icon_theme_name) ||
@@ -264,7 +256,7 @@ theme_set_custom_from_theme (const GnomeThemeMetaInfo *info, AppearanceData *dat
 
   /* select the custom theme */
   model = gtk_icon_view_get_model (icon_view);
-  if (!find_in_model (model, custom->name, COL_NAME, &iter)) {
+  if (!theme_find_in_model (model, custom->name, &iter)) {
     GtkTreeIter child;
 
     gtk_list_store_insert_with_values (data->theme_store, &child, 0,
@@ -282,67 +274,7 @@ theme_set_custom_from_theme (const GnomeThemeMetaInfo *info, AppearanceData *dat
   gtk_tree_path_free (path);
 
   /* update the theme thumbnail */
-  theme_queue_for_thumbnail (custom, data);
-}
-
-/** Theme Callbacks **/
-
-static void
-theme_changed_on_disk_cb (GnomeThemeType       type,
-			  gpointer             theme,
-			  GnomeThemeChangeType change_type,
-			  GnomeThemeElement    element,
-			  AppearanceData       *data)
-{
-  /* TODO: add/change/remove themes from the model as appropriate */
-
-  if (type == GNOME_THEME_TYPE_METATHEME) {
-    GnomeThemeMetaInfo *meta = theme;
-
-    if (change_type == GNOME_THEME_CHANGE_CREATED) {
-      gtk_list_store_insert_with_values (data->theme_store, NULL, 0,
-          COL_LABEL, meta->readable_name,
-          COL_NAME, meta->name,
-          COL_THUMBNAIL, data->theme_icon,
-          -1);
-      theme_queue_for_thumbnail (meta, data);
-
-    } else if (change_type == GNOME_THEME_CHANGE_DELETED) {
-      GtkTreeIter iter;
-
-      if (find_in_model (GTK_TREE_MODEL (data->theme_store), meta->name, COL_NAME, &iter))
-        gtk_list_store_remove (data->theme_store, &iter);
-
-    } else if (change_type == GNOME_THEME_CHANGE_CHANGED) {
-      theme_queue_for_thumbnail (meta, data);
-    }
-  }
-}
-
-static void
-theme_thumbnail_done_cb (GdkPixbuf *pixbuf, AppearanceData *data)
-{
-  GnomeThemeMetaInfo *info = data->theme_queue->data;
-
-  g_return_if_fail (info != NULL);
-
-  /* find item in model and update thumbnail */
-  if (pixbuf) {
-    GtkTreeIter iter;
-    GtkTreeModel *model = GTK_TREE_MODEL (data->theme_store);
-
-    if (find_in_model (model, info->name, COL_NAME, &iter))
-      gtk_list_store_set (data->theme_store, &iter, COL_THUMBNAIL, pixbuf, -1);
-
-    g_object_unref (pixbuf);
-  }
-
-  data->theme_queue = g_slist_remove (data->theme_queue, info);
-
-  if (data->theme_queue)
-    /* we can't call theme_thumbnail_generate directly since the thumbnail
-     * factory hasn't yet cleaned up at this point */
-    g_idle_add ((GSourceFunc) theme_thumbnail_generate, data);
+  theme_thumbnail_generate (custom, data);
 }
 
 /** GUI Callbacks **/
@@ -352,6 +284,7 @@ theme_selection_changed_cb (GtkWidget *icon_view, AppearanceData *data)
 {
   GList *selection;
   GnomeThemeMetaInfo *theme = NULL;
+  gboolean is_custom = FALSE;
 
   selection = gtk_icon_view_get_selected_items (GTK_ICON_VIEW (icon_view));
 
@@ -364,7 +297,9 @@ theme_selection_changed_cb (GtkWidget *icon_view, AppearanceData *data)
     gtk_tree_model_get_iter (model, &iter, selection->data);
     gtk_tree_model_get (model, &iter, COL_NAME, &name, -1);
 
-    if (!strcmp (name, CUSTOM_THEME_NAME))
+    is_custom = !strcmp (name, CUSTOM_THEME_NAME);
+
+    if (is_custom)
       theme = data->theme_custom;
     else
       theme = gnome_theme_meta_info_find (name);
@@ -379,6 +314,7 @@ theme_selection_changed_cb (GtkWidget *icon_view, AppearanceData *data)
 
   gtk_widget_set_sensitive (glade_xml_get_widget (data->xml, "theme_delete"),
 			    gnome_theme_is_writable (theme, GNOME_THEME_TYPE_METATHEME));
+  gtk_widget_set_sensitive (glade_xml_get_widget (data->xml, "theme_save"), is_custom);
 }
 
 static void
@@ -396,6 +332,12 @@ theme_custom_cb (GtkWidget *button, AppearanceData *data)
   gtk_widget_show_all (w);
 }
 
+static void
+theme_save_cb (GtkWidget *button, AppearanceData *data)
+{
+  theme_save_dialog_run (data->theme_custom, data);
+}
+
 static void
 theme_install_cb (GtkWidget *button, AppearanceData *data)
 {
@@ -512,6 +454,46 @@ theme_postinit (GtkIconView *icon_view, AppearanceData *data)
   g_signal_connect (gtk_settings_get_default (), "notify::gtk-color-scheme", (GCallback) theme_color_scheme_changed_cb, data);
 }
 
+static gint
+theme_list_sort_func (GnomeThemeMetaInfo *a,
+                      GnomeThemeMetaInfo *b)
+{
+  return strcmp (a->readable_name, b->readable_name);
+}
+
+static gint
+theme_store_sort_func (GtkTreeModel *model,
+                      GtkTreeIter *a,
+                      GtkTreeIter *b,
+                      gpointer user_data)
+{
+  gchar *a_name, *a_label;
+  gint rc;
+
+  gtk_tree_model_get (model, a, COL_NAME, &a_name, COL_LABEL, &a_label, -1);
+
+  if (!strcmp (a_name, CUSTOM_THEME_NAME)) {
+    rc = -1;
+  } else {
+    gchar *b_name, *b_label;
+
+    gtk_tree_model_get (model, b, COL_NAME, &b_name, COL_LABEL, &b_label, -1);
+
+    if (!strcmp (b_name, CUSTOM_THEME_NAME))
+      rc = 1;
+    else
+      rc = strcmp (a_label, b_label);
+
+    g_free (b_name);
+    g_free (b_label);
+  }
+
+  g_free (a_name);
+  g_free (a_label);
+
+  return rc;
+}
+
 void
 themes_init (AppearanceData *data)
 {
@@ -525,7 +507,7 @@ themes_init (AppearanceData *data)
   gnome_theme_init (NULL);
   gnome_wm_manager_init ();
 
-  data->theme_queue = NULL;
+  data->theme_save_dialog = NULL;
   data->theme_custom = gnome_theme_meta_info_new ();
   data->theme_icon = gdk_pixbuf_new_from_file (GNOMECC_PIXMAP_DIR "/theme-thumbnailing.png", NULL);
   data->theme_store = theme_store =
@@ -538,15 +520,12 @@ themes_init (AppearanceData *data)
   gnome_theme_info_register_theme_change ((ThemeChangedCallback) theme_changed_on_disk_cb, data);
 
   data->theme_custom->name = g_strdup (CUSTOM_THEME_NAME);
-  data->theme_custom->readable_name = g_strdup (_("Custom"));
+  data->theme_custom->readable_name = g_strdup_printf ("<i>%s</i>", _("Custom"));
   theme_load_from_gconf (data->client, data->theme_custom);
 
-
   for (l = theme_list; l; l = l->next) {
     GnomeThemeMetaInfo *info = l->data;
 
-    data->theme_queue = g_slist_prepend (data->theme_queue, info);
-
     gtk_list_store_insert_with_values (theme_store, NULL, 0,
         COL_LABEL, info->readable_name,
         COL_NAME, info->name,
@@ -556,24 +535,29 @@ themes_init (AppearanceData *data)
     if (!meta_theme && theme_is_equal (data->theme_custom, info))
       meta_theme = info;
   }
-  g_list_free (theme_list);
 
   if (!meta_theme) {
     /* add custom theme */
     meta_theme = data->theme_custom;
-    data->theme_queue = g_slist_prepend (data->theme_queue, meta_theme);
 
     gtk_list_store_insert_with_values (theme_store, NULL, 0,
         COL_LABEL, meta_theme->readable_name,
         COL_NAME, meta_theme->name,
         COL_THUMBNAIL, data->theme_icon,
         -1);
+
+    theme_thumbnail_generate (meta_theme, data);
   }
 
+  theme_list = g_list_sort (theme_list, (GCompareFunc) theme_list_sort_func);
+
+  g_list_foreach (theme_list, (GFunc) theme_thumbnail_generate, data);
+  g_list_free (theme_list);
+
   w = glade_xml_get_widget (data->xml, "theme_list");
-  gtk_icon_view_set_selection_mode (GTK_ICON_VIEW (w), GTK_SELECTION_BROWSE);
   sort_model = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (theme_store));
   gtk_icon_view_set_model (GTK_ICON_VIEW (w), GTK_TREE_MODEL (sort_model));
+  gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (sort_model), COL_LABEL, theme_store_sort_func, NULL, NULL);
   gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (sort_model), COL_LABEL, GTK_SORT_ASCENDING);
   g_signal_connect (w, "selection-changed", (GCallback) theme_selection_changed_cb, data);
   g_signal_connect_after (w, "realize", (GCallback) theme_select_name, meta_theme->name);
@@ -582,25 +566,32 @@ themes_init (AppearanceData *data)
   w = glade_xml_get_widget (data->xml, "theme_install");
   gtk_button_set_image (GTK_BUTTON (w),
                         gtk_image_new_from_stock (GTK_STOCK_OPEN, GTK_ICON_SIZE_BUTTON));
-
-  /* connect button signals */
   g_signal_connect (w, "clicked", (GCallback) theme_install_cb, data);
-  g_signal_connect (glade_xml_get_widget (data->xml, "theme_custom"), "clicked", (GCallback) theme_custom_cb, data);
+
+  w = glade_xml_get_widget (data->xml, "theme_save");
+  gtk_button_set_image (GTK_BUTTON (w),
+                        gtk_image_new_from_stock (GTK_STOCK_SAVE_AS, GTK_ICON_SIZE_BUTTON));
+  g_signal_connect (w, "clicked", (GCallback) theme_save_cb, data);
+
+  w = glade_xml_get_widget (data->xml, "theme_custom");
+  gtk_button_set_image (GTK_BUTTON (w),
+                        gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_BUTTON));
+  g_signal_connect (w, "clicked", (GCallback) theme_custom_cb, data);
+
   g_signal_connect (del_button, "clicked", (GCallback) theme_delete_cb, data);
 
   if (is_locked_down (data->client)) {
     /* FIXME: determine what needs disabling */
   }
-
-  theme_thumbnail_generate (data);
 }
 
 void
 themes_shutdown (AppearanceData *data)
 {
   gnome_theme_meta_info_free (data->theme_custom);
-  g_slist_free (data->theme_queue);
 
   if (data->theme_icon)
     g_object_unref (data->theme_icon);
+  if (data->theme_save_dialog)
+    gtk_widget_destroy (data->theme_save_dialog);
 }
