@@ -27,10 +27,10 @@
 #include <gconf/gconf-client.h>
 #include <libgnomeui/gnome-thumbnail.h>
 
+#include "gnome-theme-info.h"
+
 typedef struct {
   GConfClient *client;
-  int argc;
-  char **argv;
   GladeXML *xml;
 
   /* desktop */
@@ -50,4 +50,15 @@ typedef struct {
 
   /* font */
   GtkWidget *font_details;
+
+  /* themes */
+  GtkListStore *theme_store;
+  GnomeThemeMetaInfo *theme_custom;
+  GdkPixbuf *theme_icon;
+  GtkWidget *theme_save_dialog;
+
+  /* style */
+  GdkPixbuf *gtk_theme_icon;
+  GdkPixbuf *window_theme_icon;
+  GdkPixbuf *icon_theme_icon;
 } AppearanceData;
